import java.awt.*;

/**
 * Class ClippingVirtaulPanel: used to generate clipping polygons
 * Core data structure for rendering clip region using Weiler-Atherton algorithm.
 * Vertices on a polygon are implemented as a circular list. 
 * Each node reprepents a point, with a boolean field which indicates potential
 * starting point of a new polygon and with a ListNode field, which indicates
 * the position of intersection point on another polygon.
 * 
 */
public class ClippingVirtualPanel
{
    // inner node: data point
    private class LinkedPoints extends Point{
        ListNode brother = null;
        boolean startPoint = false;
        
        LinkedPoints(int xx, int yy){ 
            super(xx,yy); brother=null; startPoint = false;
        }
        LinkedPoints(LinkedPoints rhs){ 
            super(rhs.x,rhs.y); brother=null; startPoint = false;
        }
        LinkedPoints(Point p) { 
            super(p); brother=null; startPoint = false;
        }
    }
    
    private CircularList    source = null;      
    private CircularList    clipping = null;
    
    private Point []    src = null;
    private Point []    clp = null;
    private boolean     isOutSide = false;      // source is outside of clipping
    private boolean     isInSide = false;       // source is inside of clipping
    
    private boolean     fetching=false;     // true while extract sub polygons
    private boolean     donefetch=false;
    private ListNode    current;
    
    /**
     * Constructor: initialize two circularlist of polygons in clockwise order
     */
    public ClippingVirtualPanel(Point[] srcp, Point[] clpp){
        LinkedPoints [] points;
        
        // Constructor source polygon and clipping polygon in clockwise order.
        // Actually, as long as these two polygons are in same orientation, 
        // the algorithm is fine.
        src = srcp;     // Points[]
        points = new LinkedPoints [src.length ];
        for(int i=0;i<src.length ; i++){
            points[i] = new LinkedPoints (src[i]);
        }
        source = new CircularList ( points, Util.isClockwisePolygon (src));
        
        clp = clpp;     // Points[]
        points = new LinkedPoints [clp.length ];
        for(int i=0;i<clp.length ; i++){
            points[i] = new LinkedPoints (clp[i]);
        }
        clipping = new CircularList ( points, Util.isClockwisePolygon (clp));
    }
    
    /**
     * find the next polygon, return null if fails
     */
    public Polygon fetchNextPolygon(){
        if(donefetch){
            return null;
        } // else { System.out.println("haven't done yet"); }
        
        if(isOutSide){  // out side of clipping polygon
            // check is clipping inside of Polygon
            donefetch = true;
            if(Util.pointInPolygon(clp[0], src)){
                int [] xx = new int[clp.length];
                int [] yy = new int[clp.length];
                for (int i = 0;i<clp.length;i++){
                    xx[i] = clp[i].x;
                    yy[i] = clp[i].y;
                }
                return new Polygon(xx,yy,clp.length);
            } else {
                // System.out.println("source is totally outside");
                return null;
            }
        } // else { System.out.println(" interleaving with each other"); }
        
        if(isInSide){
            // return source polygon
            // System.out.println("Source is inside of clipping polygon");
            int [] xx = new int[src.length];
            int [] yy = new int[src.length];
            for (int i = 0;i<src.length;i++){
                xx[i] = src[i].x;
                yy[i] = src[i].y;
            }
            donefetch = true;
            return new Polygon(xx,yy,src.length);
        }
        
        boolean gotone = false;
        if(fetching==false) {   // start to search
            // System.out.println("first fetch");
            fetching = true;
            source.goToTheBeginning();
        } else {
            // System.out.println("following fetch");
            source.setCurrentPosition (current);  // restore the previous session
        }
        
        LinkedPoints currp;
        do{ // find the next starting point
            // System.out.println("do while loop");
            currp =(LinkedPoints)source.getCurrentObj();
            if(currp.startPoint == true){
                current = source.getCurrentPosition();
                gotone = true;
                break;
            }
        } while(source.goToTheNext());
        
        if(gotone==false){
            // System.out.println("didn't find a starting point, no you bed it");
            donefetch = true;
            return null;
        } else {
            // System.out.println("find a starting point");
            Polygon p = new Polygon();
            ListNode start = source.getCurrentPosition();
            LinkedPoints startp = (LinkedPoints)start.getData();
            
            startp.startPoint = false;
            
            ListNode travel = start.getNextNode ();
            LinkedPoints travelp;
            
            p.addPoint(startp.x, startp.y);
            while(true){
                travelp = (LinkedPoints)travel.getData ();
                p.addPoint(travelp.x, travelp.y);
                travelp.startPoint = false;
                if( travelp.brother != null ){  // switch to another list
                        travel = travelp.brother;
                        if(travel==start) break;
                        travel = travel.getNextNode ();
                } else {    // move on
                    travel = travel.getNextNode ();
                }
            }
            return p;
        }
    }

    
    /**
     * Find the starting point of source polygon: it has to be one of the
     * vertices out side of the clipping polygon.
     * Rearrage the source polygon at the same time.
     */
    public boolean findStartPoint(){
        source.goToTheBeginning();
        do{
            Point p = (Point)source.getCurrentObj();
            if(!Util.pointInPolygon((Point) source.getCurrentObj(), clp)){
                source.resetStartingToCurrent();
                return true;
            }
        } while (source.goToTheNext());
        isInSide = true;
        return false;
    }
    
    /**
     * Find all the vertices on the intersections
     */
    public void findIntersections(){
        // if source polygon is out inside of clipping polygon, done
        LinkedPoints currp, nextp;     // two contineous points on source polygon
        
        boolean IsCurrentInside, IsNextInside;
        if(findStartPoint()){       
            isOutSide = true;
            source.goToTheBeginning();
            float [] t = new float[2];
            do{
                currp = (LinkedPoints)source.getCurrentObj();
                IsCurrentInside = Util.pointInPolygon(currp, clp);
                nextp = (LinkedPoints)source.getTheNextObj();
                IsNextInside = Util.pointInPolygon(nextp, clp);
                // find the intersection points in the increasing order of t
                LinkedPoints[] intersects = new 
                                  LinkedPoints [clipping.getSize()];
                LinkedPoints    PointOnClip;
                float [] keys = new float[clipping.getSize()];
                clipping.goToTheBeginning ();
                LinkedPoints p1onclip;
                LinkedPoints p2onclip;
                int index = 0;
                do{ // for eche line segment on source polygon, finds all the
                    // intersection points on the clipping polygon and stores
                    // them in sorted way
                    p1onclip = (LinkedPoints)clipping.getCurrentObj();
                    p2onclip = (LinkedPoints)clipping.getTheNextObj();
                    if(Util.isTwoLineIntersect(currp, nextp, p1onclip, 
                                               p2onclip, t)){
                        intersects[index] =  new LinkedPoints(
                                 (int)((1-t[0])*currp.x+t[0]*nextp.x),
                                 (int)((1-t[0])*currp.y+t[0]*nextp.y));
                        PointOnClip = new LinkedPoints(intersects[index]);
                        clipping.insertAfterCurrent(PointOnClip);
                        clipping.goToTheNext ();
                        intersects[index].brother = clipping.getCurrentPosition();
                        keys[index++] = t[0];
                    }
                } while(clipping.goToTheNext());
                
                // now have the intersection points list in intersects[] and keys
                // inserts into the source polygon
                if(index>0){
                    isOutSide = false;
                    float [] newkeys = new float[index];
                    int [] sorted_index = new int[index];
                    for(int i=0;i<index;i++){
                        sorted_index[i] = i;
                        newkeys[i] = keys[i];
                    }
                    Sort.sort(sorted_index, newkeys);
                    
                    for(int i=0;i<index;i++){ // insert into source polygon
                        int sindex = sorted_index[i];
                        if(IsCurrentInside){
                            IsCurrentInside = false; // refer to the next
                        } else {
                            IsCurrentInside = true; // refer to the next, switch
                            intersects[sindex].startPoint = true;
                        }
                        source.insertAfterCurrent(intersects[sindex]);
                        source.goToTheNext();
                        ((LinkedPoints)intersects[sindex].brother.getData()).brother
                            = source.getCurrentPosition();
                    }
                } // esle:  no intersect, move on
                // currp = nextp;
                // IsCurrentInside = IsNextInside;
            } while(source.goToTheNext());
        } 
        
        // check source and clipping :: debugging
        /***********************************************************************
        System.out.println("isOutSide: " + isOutSide + ", isInSide: " + isInSide +
                           ", donefetch: " + donefetch + ", fetching? " + fetching);
        
        System.out.println("Checkout the source and clipping");
        source.goToTheBeginning ();
        do{
            currp = (LinkedPoints)source.getCurrentObj();
            System.out.println("src: "+ currp.x + ", " + currp.y +
                               ", p1.st=null "+ (currp.brother == null)+
                               ", is strating points? "+ currp.startPoint);
            if(currp.brother != null){
                if(((LinkedPoints)(currp.brother.getData())).brother.getData()
                                    == currp){
                    System.out.println("great link");
                } else {
                    System.out.print("damn it");
                }
            }
        } while(source.goToTheNext ());
        clipping.goToTheBeginning ();
        do{
            currp = (LinkedPoints)clipping.getCurrentObj();
            System.out.println("clp: "+ currp.x + ", " + currp.y +
                               ", p1.st=null "+ (currp.brother == null)+
                               ", is strating points? " + currp.startPoint);
            if(currp.brother != null){
                if(((LinkedPoints)(currp.brother.getData())).brother.getData()
                                == currp){
                    System.out.println("great link");
                } else {
                    System.out.print("damn it");
                }
            }
        } while(clipping.goToTheNext());
        ********************************************************************/
    }
    
    // test driver
    public static void main(String args[]){
        Point[] src = new Point[3];
        Point[] clp = new Point[4];
        
        src[0]=new Point(0,0);
        src[1]=new Point(100,0);
        src[2]=new Point(0,100);
        clp[0]=new Point(10,10);
        clp[1]=new Point(100,10);
        clp[2]=new Point(100,100);
        clp[3]=new Point(10,100);

        ClippingVirtualPanel panel = new ClippingVirtualPanel(src, clp);
        panel.findIntersections ();
        Polygon p = panel.fetchNextPolygon ();
        while(p!=null){
            p = panel.fetchNextPolygon ();
        }
    }
}
